from __future__ import print_function

from distutils.version import LooseVersion
from logging import getLogger
import select
import sys
from threading import Thread
from time import sleep
import webbrowser

import wx

from .._wxutils import ID, Icon
from ..plot._base import CONFIG
from .about import AboutFrame


APP = None  # hold the App instance
IS_OSX = sys.platform == 'darwin'
IS_WINDOWS = sys.platform.startswith('win')


def wildcard(filetypes):
    if filetypes:
        return '|'.join(map('|'.join, filetypes))
    else:
        return ""


class App(wx.App):
    def OnInit(self):
        self.SetAppName("Eelbrain")
        self.SetAppDisplayName("Eelbrain")

        # File Menu
        m = file_menu = wx.Menu()
        m.Append(wx.ID_OPEN, '&Open... \tCtrl+O')
        m.AppendSeparator()
        m.Append(wx.ID_CLOSE, '&Close Window \tCtrl+W')
        m.Append(wx.ID_SAVE, "Save \tCtrl+S")
        m.Append(wx.ID_SAVEAS, "Save As... \tCtrl+Shift+S")

        # Edit Menu
        m = edit_menu = wx.Menu()
        m.Append(ID.UNDO, '&Undo \tCtrl+Z')
        m.Append(ID.REDO, '&Redo \tCtrl+Shift+Z')
        m.AppendSeparator()
        m.Append(wx.ID_CUT, 'Cut \tCtrl+X')
        m.Append(wx.ID_COPY, 'Copy \tCtrl+C')
        m.Append(ID.COPY_AS_PNG, 'Copy as PNG \tCtrl+Shift+C')
        m.Append(wx.ID_PASTE, 'Paste \tCtrl+V')
        m.AppendSeparator()
        m.Append(wx.ID_CLEAR, 'Cle&ar')

        # Tools Menu
        # updated by the active GUI
        tools_menu = wx.Menu()

        # View Menu
        m = view_menu = wx.Menu()
        m.Append(ID.SET_VLIM, "Set Y-Axis Limit... \tCtrl+l", "Change the Y-"
                 "axis limit in epoch plots")
        m.Append(ID.SET_MARKED_CHANNELS, "Mark Channels...", "Mark specific "
                 "channels in plots")
        self._draw_crosshairs_menu_item = m.Append(
            ID.DRAW_CROSSHAIRS, "Draw &Crosshairs", "Draw crosshairs under the "
            "cursor", kind=wx.ITEM_CHECK)
        m.AppendSeparator()
        m.Append(ID.SET_LAYOUT, "&Set Layout... \tCtrl+Shift+l", "Change the "
                 "page layout")

        # Go Menu
        m = go_menu = wx.Menu()
        m.Append(wx.ID_FORWARD, '&Forward \tCtrl+]', 'Go One Page Forward')
        m.Append(wx.ID_BACKWARD, '&Back \tCtrl+[', 'Go One Page Back')

        # Window Menu
        m = window_menu = wx.Menu()
        m.Append(ID.WINDOW_MINIMIZE, '&Minimize \tCtrl+M')
        m.Append(ID.WINDOW_ZOOM, '&Zoom')
        m.AppendSeparator()
        m.Append(ID.WINDOW_TILE, '&Tile')
        m.AppendSeparator()
        self.window_menu_window_items = []

        # Help Menu
        m = help_menu = wx.Menu()
        m.Append(ID.HELP_EELBRAIN, 'Eelbrain Help')
        m.Append(ID.HELP_PYTHON, "Python Help")
        m.AppendSeparator()
        m.Append(wx.ID_ABOUT, '&About Eelbrain')

        # Menu Bar
        menu_bar = wx.MenuBar()
        menu_bar.Append(file_menu, "File")
        menu_bar.Append(edit_menu, "Edit")
        menu_bar.Append(tools_menu, "Tools")
        menu_bar.Append(view_menu, "View")
        menu_bar.Append(go_menu, "Go")
        menu_bar.Append(window_menu, "Window")
        menu_bar.Append(help_menu, self.GetMacHelpMenuTitleName() if IS_OSX else 'Help')
        wx.MenuBar.MacSetCommonMenuBar(menu_bar)
        self.menubar = menu_bar

        # Dock icon
        self.dock_icon = DockIcon(self)

        # Bind Menu Commands
        self.Bind(wx.EVT_MENU_OPEN, self.OnMenuOpened)
        self.Bind(wx.EVT_MENU, self.OnAbout, id=wx.ID_ABOUT)
        self.Bind(wx.EVT_MENU, self.OnOpen, id=wx.ID_OPEN)
        self.Bind(wx.EVT_MENU, self.OnClear, id=wx.ID_CLEAR)
        self.Bind(wx.EVT_MENU, self.OnCloseWindow, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_MENU, self.OnCopy, id=wx.ID_COPY)
        self.Bind(wx.EVT_MENU, self.OnCopyAsPNG, id=ID.COPY_AS_PNG)
        self.Bind(wx.EVT_MENU, self.OnCut, id=wx.ID_CUT)
        self.Bind(wx.EVT_MENU, self.OnDrawCrosshairs, id=ID.DRAW_CROSSHAIRS)
        self.Bind(wx.EVT_MENU, self.OnOnlineHelp, id=ID.HELP_EELBRAIN)
        self.Bind(wx.EVT_MENU, self.OnOnlineHelp, id=ID.HELP_PYTHON)
        self.Bind(wx.EVT_MENU, self.OnPaste, id=wx.ID_PASTE)
        self.Bind(wx.EVT_MENU, self.OnRedo, id=ID.REDO)
        self.Bind(wx.EVT_MENU, self.OnSave, id=wx.ID_SAVE)
        self.Bind(wx.EVT_MENU, self.OnSaveAs, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_MENU, self.OnSetLayout, id=ID.SET_LAYOUT)
        self.Bind(wx.EVT_MENU, self.OnSetMarkedChannels, id=ID.SET_MARKED_CHANNELS)
        self.Bind(wx.EVT_MENU, self.OnSetVLim, id=ID.SET_VLIM)
        self.Bind(wx.EVT_MENU, self.OnUndo, id=ID.UNDO)
        self.Bind(wx.EVT_MENU, self.OnWindowMinimize, id=ID.WINDOW_MINIMIZE)
        self.Bind(wx.EVT_MENU, self.OnWindowTile, id=ID.WINDOW_TILE)
        self.Bind(wx.EVT_MENU, self.OnWindowZoom, id=ID.WINDOW_ZOOM)
        self.Bind(wx.EVT_MENU, self.OnQuit, id=wx.ID_EXIT)
        self.Bind(wx.EVT_MENU, self.OnYieldToTerminal, id=ID.YIELD_TO_TERMINAL)

        # bind update UI
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIBackward, id=wx.ID_BACKWARD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIClear, id=wx.ID_CLEAR)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIClose, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUICopy, id=wx.ID_COPY)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUICopyAsPNG, id=ID.COPY_AS_PNG)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUICut, id=wx.ID_CUT)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIDown, id=wx.ID_DOWN)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIDrawCrosshairs, id=ID.DRAW_CROSSHAIRS)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIForward, id=wx.ID_FORWARD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIOpen, id=wx.ID_OPEN)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIPaste, id=wx.ID_PASTE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIRedo, id=ID.REDO)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISave, id=wx.ID_SAVE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISaveAs, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISetLayout, id=ID.SET_LAYOUT)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISetMarkedChannels, id=ID.SET_MARKED_CHANNELS)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISetVLim, id=ID.SET_VLIM)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUITools, id=ID.TOOLS)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIUndo, id=ID.UNDO)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIUp, id=wx.ID_UP)

        # register in IPython
        self.using_prompt_toolkit = False
        self._ipython = None
        if ('IPython' in sys.modules and
                LooseVersion(sys.modules['IPython'].__version__) >=
                LooseVersion('5') and CONFIG['prompt_toolkit']):
            import IPython

            IPython.terminal.pt_inputhooks.register('eelbrain',
                                                    self.pt_inputhook)
            shell = IPython.get_ipython()
            if shell is not None:
                try:
                    shell.enable_gui('eelbrain')
                except IPython.core.error.UsageError:
                    print("Prompt-toolkit does not seem to be supported by "
                          "the current IPython shell (%s); The Eelbrain GUI "
                          "needs to block Terminal input to work. Use "
                          "eelbrain.gui.run() to start GUI interaction." %
                          shell.__class__.__name__)
                else:
                    self.using_prompt_toolkit = True
                    self._ipython = shell

        self.SetExitOnFrameDelete(not self.using_prompt_toolkit)
        if not self.using_prompt_toolkit:
            go_menu.AppendSeparator()
            go_menu.Append(ID.YIELD_TO_TERMINAL, '&Yield to Terminal \tAlt+Ctrl+Q')


        return True

    def pt_inputhook(self, context):
        """prompt_toolkit inputhook"""
        # prompt_toolkit.eventloop.inputhook.InputHookContext
        if IS_WINDOWS:
            # On Windows, select.poll() is not available
            def thread():
                while not context.input_is_ready():
                    sleep(.02)
                wx.CallAfter(self.pt_yield)
        else:
            def thread():
                poll = select.poll()
                poll.register(context.fileno(), select.POLLIN)
                poll.poll(-1)
                wx.CallAfter(self.pt_yield)

        Thread(target=thread).start()
        self.MainLoop()

    def pt_yield(self):
        self.ExitMainLoop(True)

    def _get_active_frame(self):
        win = wx.Window.FindFocus()
        win_parent = wx.GetTopLevelParent(win)
        if win_parent:
            return win_parent
        for w in wx.GetTopLevelWindows():
            if hasattr(w, 'IsActive') and w.IsActive():
                return w
        return wx.GetActiveWindow()

    def _get_parent_gui(self):
        frame = self._get_active_frame()
        if frame is None:
            return
        while True:
            if hasattr(frame, 'MakeToolsMenu'):
                return frame
            elif frame.Parent is not None:
                frame = frame.Parent
            else:
                return

    def _bash_ui(self, func, *args):
        "Launch a modal dialog based on terminal input"
        if self.using_prompt_toolkit or self.IsMainLoopRunning():
            return func(*args)
        else:
            if not self.GetTopWindow():
                self.SetTopWindow(wx.Frame(None))
            wx.CallLater(10, func, *args)
            print("Please switch to the Python Application to provide input.")
            self.MainLoop()
            return self._result

    def _bash_ui_finalize(self, result):
        if self.using_prompt_toolkit or self.IsMainLoopRunning():
            return result
        else:
            self._result = result
            self.ExitMainLoop()

    def ask_for_dir(self, title="Select Folder", message="Please Pick a Folder",
                    must_exist=True):
        return self._bash_ui(self._ask_for_dir, title, message, must_exist)

    def _ask_for_dir(self, title, message, must_exist):
        style = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
        if must_exist:
            style = style | wx.DD_DIR_MUST_EXIST

        dialog = wx.DirDialog(None, message, name=title,
                              style=style)
        dialog.SetTitle(title)
        if dialog.ShowModal() == wx.ID_OK:
            result = dialog.GetPath()
        else:
            result = False
        dialog.Destroy()
        return self._bash_ui_finalize(result)

    def ask_for_file(self, title, message, filetypes, directory, mult):
        return self._bash_ui(self._ask_for_file, title, message, filetypes,
                             directory, mult)

    def _ask_for_file(self, title, message, filetypes, directory, mult):
        """Return path(s) or False.

        Parameters
        ----------
        ...
        directory : str
            Path to initial directory.

        Returns
        -------
        Result : list | str | None
            Paths(s) or False.
        """
        style = wx.FD_OPEN
        if mult:
            style = style | wx.FD_MULTIPLE
        dialog = wx.FileDialog(None, message, directory,
                               wildcard=wildcard(filetypes), style=style)
        dialog.SetTitle(title)
        if dialog.ShowModal() == wx.ID_OK:
            if mult:
                result = dialog.GetPaths()
            else:
                result = dialog.GetPath()
        else:
            result = False
        dialog.Destroy()
        return self._bash_ui_finalize(result)

    def ask_for_string(self, title, message, default='', parent=None):
        return self._bash_ui(self._ask_for_string, title, message, default,
                             parent)

    def _ask_for_string(self, title, message, default, parent):
        dialog = wx.TextEntryDialog(parent, message, title, default)
        if dialog.ShowModal() == wx.ID_OK:
            result = dialog.GetValue()
        else:
            result = False
        dialog.Destroy()
        return self._bash_ui_finalize(result)

    def ask_saveas(self, title, message, filetypes, defaultDir, defaultFile):
        return self._bash_ui(self._ask_saveas, title, message, filetypes,
                             defaultDir, defaultFile)

    def _ask_saveas(self, title, message, filetypes, defaultDir, defaultFile):
        # setup file-dialog
        dialog = wx.FileDialog(None, message, wildcard=wildcard(filetypes),
                               style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        dialog.SetTitle(title)
        if defaultDir:
            dialog.SetDirectory(defaultDir)
        if defaultFile:
            dialog.SetFilename(defaultFile)
        # get result
        if dialog.ShowModal() == wx.ID_OK:
            result = dialog.GetPath()
        else:
            result = False
        dialog.Destroy()
        return self._bash_ui_finalize(result)

    def message_box(self, message, caption, style, parent=None):
        return self._bash_ui(self._message_box, message, caption, style, parent)

    def _message_box(self, message, caption, style, parent):
        dialog = wx.MessageDialog(parent, message, caption, style)
        result = dialog.ShowModal()
        dialog.Destroy()
        self._bash_ui_finalize(result)

    def ExitMainLoop(self, event_with_pt=True):
        if event_with_pt or not self.using_prompt_toolkit:
            # with prompt-toolkit, this leads to hanging when terminating the
            # interpreter
            wx.App.ExitMainLoop(self)

    def Attach(self, obj, desc, default_name, parent):
        if self._ipython is None:
            self.message_box(
                "Attach Unavailable",
                "The attach command requires running from within IPython 5 or "
                "later", wx.ICON_ERROR|wx.OK, parent)
            return
        name = self.ask_for_string(
            "Attach", "Variable name for %s in terminal:" % desc, default_name,
            parent)
        if name:
            self._ipython.user_global_ns[name] = obj

    def OnAbout(self, event):
        if hasattr(self, '_about_frame') and hasattr(self._about_frame, 'Raise'):
            self._about_frame.Raise()
        else:
            self._about_frame = AboutFrame(None)
            self._about_frame.Show()
#             frame.SetFocus()

    def OnClear(self, event):
        frame = self._get_active_frame()
        frame.OnClear(event)

    def OnCloseWindow(self, event):
        frame = self._get_active_frame()
        if frame is not None:
            frame.Close()

    def OnCopy(self, event):
        win = wx.Window.FindFocus()
        if hasattr(win, 'CanCopy'):
            return win.Copy()
        win = self._get_active_frame()
        if hasattr(win, 'CanCopy'):
            return win.Copy()
        getLogger('Eelbrain').debug(
            "App.OnCopy() call but neither focus nor frame have CanCopy()")
        event.Skip()

    def OnCopyAsPNG(self, event):
        wx.Window.FindFocus().CopyAsPNG()
#
    def OnCut(self, event):
        win = wx.Window.FindFocus()
        win.Cut()

    def OnDrawCrosshairs(self, event):
        frame = self._get_active_frame()
        if hasattr(frame, 'SetDrawCrosshairs'):
            frame.SetDrawCrosshairs(self._draw_crosshairs_menu_item.IsChecked())

    def OnMenuOpened(self, event):
        "Update window names in the window menu"
        menu = event.GetMenu()
        if menu.GetTitle() == 'Window':
            # clear old entries
            while self.window_menu_window_items:
                item = self.window_menu_window_items.pop()
                menu.RemoveItem(item)
                self.Unbind(wx.EVT_MENU, id=item.GetId())
            # add new entries
            for window in wx.GetTopLevelWindows():
                id_ = window.GetId()
                label = window.GetTitle()
                if not label:
                    continue
                item = menu.Append(id_, label)
                self.Bind(wx.EVT_MENU, self.OnWindowRaise, id=id_)
                self.window_menu_window_items.append(item)
        elif menu.GetTitle() == 'Tools':
            for item in menu.GetMenuItems():
                menu.RemoveItem(item)
            frame = self._get_parent_gui()
            if frame:
                frame.MakeToolsMenu(menu)

    def OnOnlineHelp(self, event):
        "Called from the Help menu to open external resources"
        Id = event.GetId()
        if Id == ID.HELP_EELBRAIN:
            webbrowser.open("https://pythonhosted.org/eelbrain/")
        elif Id == ID.HELP_PYTHON:
            webbrowser.open("http://docs.python.org/2.7/")
        else:
            raise RuntimeError("Invalid help ID")

    def OnOpen(self, event):
        frame = self._get_active_frame()
        frame.OnOpen(event)

    def OnPaste(self, event):
        win = wx.Window.FindFocus()
        win.Paste()

    def OnQuit(self, event):
        for win in wx.GetTopLevelWindows():
            if not win.Close():
                return
        if not self.using_prompt_toolkit:
            self.ExitMainLoop()

    def OnRedo(self, event):
        frame = self._get_active_frame()
        frame.OnRedo(event)

    def OnSave(self, event):
        frame = self._get_active_frame()
        frame.OnSave(event)

    def OnSaveAs(self, event):
        frame = self._get_active_frame()
        frame.OnSaveAs(event)

    def OnSetVLim(self, event):
        frame = self._get_active_frame()
        frame.OnSetVLim(event)

    def OnSetLayout(self, event):
        frame = self._get_active_frame()
        frame.OnSetLayout(event)

    def OnSetMarkedChannels(self, event):
        frame = self._get_active_frame()
        frame.OnSetMarkedChannels(event)

    def OnUndo(self, event):
        frame = self._get_active_frame()
        frame.OnUndo(event)

    def OnUpdateUIBackward(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIBackward'):
            frame.OnUpdateUIBackward(event)
        else:
            event.Enable(False)

    def OnUpdateUIClear(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIClear'):
            frame.OnUpdateUIClear(event)
        else:
            event.Enable(False)

    def OnUpdateUIClose(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIClose'):
            frame.OnUpdateUIClose(event)
        else:
            event.Enable(False)

    def OnUpdateUICopy(self, event):
        win = wx.Window.FindFocus()
        if hasattr(win, 'CanCopy'):
            return event.Enable(win.CanCopy())
        win = self._get_active_frame()
        if hasattr(win, 'CanCopy'):
            return event.Enable(win.CanCopy())

    def OnUpdateUICopyAsPNG(self, event):
        event.Enable(hasattr(wx.Window.FindFocus(), 'CopyAsPNG'))

    def OnUpdateUICut(self, event):
        win = wx.Window.FindFocus()
        event.Enable(win and hasattr(win, 'CanCut') and win.CanCut())

    def OnUpdateUIDown(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIDown'):
            frame.OnUpdateUIDown(event)
        else:
            event.Enable(False)

    def OnUpdateUIDrawCrosshairs(self, event):
        frame = self._get_active_frame()
        valid_frame = hasattr(frame, 'SetDrawCrosshairs')
        event.Enable(valid_frame)
        event.Check(frame.GetDrawCrosshairs() if valid_frame else False)

    def OnUpdateUIForward(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIForward'):
            frame.OnUpdateUIForward(event)
        else:
            event.Enable(False)

    def OnUpdateUIOpen(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIOpen'):
            frame.OnUpdateUIOpen(event)
        else:
            event.Enable(False)

    def OnUpdateUIPaste(self, event):
        win = wx.Window.FindFocus()
        event.Enable(win and hasattr(win, 'CanPaste') and win.CanPaste())

    def OnUpdateUIRedo(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIRedo'):
            frame.OnUpdateUIRedo(event)
        else:
            event.Enable(False)

    def OnUpdateUISave(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISave'):
            frame.OnUpdateUISave(event)
        else:
            event.Enable(False)

    def OnUpdateUISaveAs(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISaveAs'):
            frame.OnUpdateUISaveAs(event)
        else:
            event.Enable(False)

    def OnUpdateUISetLayout(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISetLayout'):
            frame.OnUpdateUISetLayout(event)
        else:
            event.Enable(False)

    def OnUpdateUISetMarkedChannels(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISetMarkedChannels'):
            frame.OnUpdateUISetMarkedChannels(event)
        else:
            event.Enable(False)

    def OnUpdateUISetVLim(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUISetVLim'):
            frame.OnUpdateUISetVLim(event)
        else:
            event.Enable(False)

    def OnUpdateUITools(self, event):
        event.Enable(bool(self._get_parent_gui))

    def OnUpdateUIUndo(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIUndo'):
            frame.OnUpdateUIUndo(event)
        else:
            event.Enable(False)

    def OnUpdateUIUp(self, event):
        frame = self._get_active_frame()
        if frame and hasattr(frame, 'OnUpdateUIUp'):
            frame.OnUpdateUIUp(event)
        else:
            event.Enable(False)

    def OnWindowMinimize(self, event):
        frame = self._get_active_frame()
        if frame:
            frame.Iconize()

    def OnWindowRaise(self, event):
        id_ = event.GetId()
        window = wx.FindWindowById(id_)
        window.Raise()

    def OnWindowTile(self, event):
        frames = sorted(wx.GetTopLevelWindows(), lambda x, y: x.Position[0] < y.Position[0])
        dx, dy = wx.DisplaySize()
        x = 0
        y = 0
        y_next = 0
        for frame in frames:
            sx, sy = frame.Size
            if x and x + sx > dx:
                if y_next > dy:
                    return
                x = 0
                y = y_next
            frame.Position = (x, y)
            y_next = max(y_next, y + sy)
            x += sx

    def OnWindowZoom(self, event):
        frame = self._get_active_frame()
        if frame:
            frame.Maximize()

    def OnYieldToTerminal(self, event):
        self.ExitMainLoop()


def get_app():
    global APP
    if APP is None:
        try:
            APP = App()
        except SystemExit as exc:
            if exc.code.startswith("This program needs access to the screen"):
                raise SystemExit(
                    exc.code +
                    "\n\nTo make sure you are running a framework build, "
                    "launch IPython with:\n\n    $ eelbrain")
            else:
                raise

    return APP


def run(block=False):
    app = get_app()
    if app.using_prompt_toolkit:
        if block:
            app.MainLoop()
    else:
        if not app.IsMainLoopRunning():
            print("Starting GUI. Quit the Python application to return to the "
                  "shell...")
            app.MainLoop()


class DockIcon(wx.TaskBarIcon):
    # http://stackoverflow.com/a/38249390/166700
    def __init__(self, app):
        wx.TaskBarIcon.__init__(self, iconType=wx.TBI_DOCK)
        self.app = app

        # Set the image
        self.SetIcon(Icon('eelbrain256', True), "Eelbrain")
        self.imgidx = 1

    def CreatePopupMenu(self):
        if not self.app.using_prompt_toolkit:
            menu = wx.Menu()
            menu.Append(ID.YIELD_TO_TERMINAL, '&Yield to Terminal')
            return menu



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import os

from nose.tools import eq_, ok_, assert_raises
import numpy as np
from numpy.testing import assert_array_equal

from eelbrain import Dataset, Factor, Var, MneExperiment
from eelbrain._utils.testing import assert_dataobj_equal, TempDir


SUBJECT = 'CheeseMonger'
SUBJECTS = ['R%04i' % i for i in (1, 11, 111, 1111)]
SAMPLINGRATE = 1000.
TRIGGERS = np.tile(np.arange(1, 5), 2)
I_START = np.arange(1001, 1441, 55)


class BaseExperiment(MneExperiment):

    path_version = 1

    sessions = 'file'


class EventExperiment(MneExperiment):

    path_version = 1

    trigger_shift = 0.03

    sessions = 'cheese'

    variables = {'kind': {(1, 2, 3, 4): 'cheese', (11, 12, 13, 14): 'pet'},
                 'name': {1: 'Leicester', 2: 'Tilsit', 3: 'Caerphilly',
                          4: 'Bel Paese'},
                 'backorder': {(1, 4): 'no', (2, 3): 'yes'},
                 'taste': {(1, 2): 'good', 'default': 'bad'}}

    epochs = {'cheese': {'sel': "kind == 'cheese'",
                         'tmin': -0.2},
              'cheese-leicester': {'base': 'cheese',
                                   'tmin': -0.1,
                                   'sel': "name == 'Leicester'"},
              'cheese-tilsit': {'base': 'cheese',
                                'sel': "name == 'Tilsit"}}

    defaults = {'model': 'name'}


class EventExperimentTriggerShiftDict(EventExperiment):
    "Test trigger shift as dictionary"
    trigger_shift = {SUBJECT: 0.04}


def gen_triggers():
    raw = Var([], info={'sfreq': SAMPLINGRATE})
    ds = Dataset(info={'subject': SUBJECT, 'raw': raw, 'sfreq': SAMPLINGRATE})
    ds['trigger'] = Var(TRIGGERS)
    ds['i_start'] = Var(I_START)
    return ds


def test_mne_experiment_templates():
    "Test MneExperiment template formatting"
    tempdir = TempDir()
    e = BaseExperiment(tempdir, False)

    # Don't create dirs without root
    ok_(e.get('raw-file', mkdir=True).endswith('-raw.fif'))

    # model
    eq_(e.get('model', model='a % b'), 'a%b')
    eq_(e.get('model', model='b % a'), 'a%b')
    assert_raises(ValueError, e.set, model='a*b')
    assert_raises(ValueError, e.set, model='log(a)')

    # compounds
    eq_(e.get('src_kind'), '0-40 bestreg free-3-dSPM')
    e.set_inv('fixed')
    eq_(e.get('src_kind'), '0-40 bestreg fixed-3-dSPM')
    e.set(cov='noreg')
    eq_(e.get('src_kind'), '0-40 noreg fixed-3-dSPM')
    e.set(raw='1-40')
    eq_(e.get('src_kind'), '1-40 noreg fixed-3-dSPM')
    e.set(src='ico-5')
    eq_(e.get('src_kind'), '1-40 noreg ico-5 fixed-3-dSPM')
    e.set(src='ico-4')
    eq_(e.get('src_kind'), '1-40 noreg fixed-3-dSPM')

    # find terminal field names
    eq_(e.find_keys('raw-file'), {'root', 'subject', 'session'})
    eq_(e.find_keys('evoked-file', False),
        {'subject', 'session', 'modality', 'raw', 'epoch', 'rej',
         'equalize_evoked_count', 'model', })

    # inv
    SNR1 = 1.
    SNR2 = 1. / 2**2
    SNR3 = 1. / 3**2

    def set_inv(inv, make_kw, apply_kw, *args):
        e.reset()
        e.set(inv=inv)
        eq_(e._params['make_inv_kw'], make_kw)
        eq_(e._params['apply_inv_kw'], apply_kw)
        e.reset()
        e.set_inv(*args)
        eq_(e.get('inv'), inv)
        eq_(e._params['make_inv_kw'], make_kw)
        eq_(e._params['apply_inv_kw'], apply_kw)

    yield (set_inv, 'free-3-MNE',
           {'loose': 1, 'depth': 0.8},
           {'method': 'MNE', 'lambda2': SNR3},
           'free', 3, 'MNE')
    yield (set_inv, 'free-3-dSPM-0.2-pick_normal',
           {'loose': 1, 'depth': 0.2},
           {'method': 'dSPM', 'lambda2': SNR3, 'pick_normal': True},
           'free', 3, 'dSPM', .2, True)
    yield (set_inv, 'fixed-2-MNE-0.2',
           {'fixed': True, 'loose': None, 'depth': 0.2},
           {'method': 'MNE', 'lambda2': SNR2},
           'fixed', 2, 'MNE', .2)
    yield (set_inv, 'fixed-2-MNE-pick_normal',
           {'fixed': True, 'loose': None, 'depth': 0.8},
           {'method': 'MNE', 'lambda2': SNR2, 'pick_normal': True},
           'fixed', 2, 'MNE', None, True)
    yield (set_inv, 'loose.5-3-sLORETA',
           {'loose': 0.5, 'depth': 0.8},
           {'method': 'sLORETA', 'lambda2': SNR3},
           0.5, 3, 'sLORETA')
    yield (set_inv, 'fixed-1-MNE-0',
           {'fixed': True, 'loose': None, 'depth': None},
           {'method': 'MNE', 'lambda2': SNR1},
           'fixed', 1, 'MNE', 0)
    # should remove this
    yield (set_inv, 'fixed-1-MNE-0.8',
           {'fixed': True, 'loose': None, 'depth': 0.8},
           {'method': 'MNE', 'lambda2': SNR1},
           'fixed', 1, 'MNE', 0.8)

    assert_raises(ValueError, e.set_inv, 'free', -3, 'dSPM')
    assert_raises(ValueError, e.set, inv='free-3-mne')
    assert_raises(ValueError, e.set, inv='free-3-MNE-2')


def test_test_experiment():
    "Test event labeling with the EventExperiment subclass of MneExperiment"
    e = EventExperiment()

    # test defaults
    eq_(e.get('session'), 'cheese')
    eq_(e.get('model'), 'name')

    # test event labeling
    ds = e.label_events(gen_triggers())
    name = Factor([e.variables['name'][t] for t in TRIGGERS], name='name')
    assert_dataobj_equal(ds['name'], name)
    tgt = ds['trigger'].as_factor(e.variables['backorder'], 'backorder')
    assert_dataobj_equal(ds['backorder'], tgt)
    tgt = ds['trigger'].as_factor(e.variables['taste'], 'taste')
    assert_dataobj_equal(ds['taste'], tgt)
    assert_array_equal(ds['i_start'], I_START)
    assert_array_equal(ds['subject'] == SUBJECT, True)

    # tests disabled (trigger-shift applied in load_events):
    # ---
    #  assert_equal(ds['i_start'], I_START + round(0.03 * SAMPLINGRATE))
    # # test without trigger shift
    # e.trigger_shift = 0
    # ds = e.label_events(gen_triggers())
    # assert_equal(ds['i_start'], I_START)
    # # trigger shift dict
    # e2 = EventExperimentTriggerShiftDict('', False)
    # ds = e2.label_events(gen_triggers())
    # assert_equal(ds['i_start'], I_START + round(0.04 * SAMPLINGRATE))

    # epochs
    eq_(e._epochs['cheese'].tmin, -0.2)
    eq_(e._epochs['cheese-leicester'].tmin, -0.1)
    eq_(e._epochs['cheese-tilsit'].tmin, -0.2)


class FileExperiment(MneExperiment):

    path_version = 1

    auto_delete_cache = 'disable'

    groups = {'gsub': SUBJECTS[1:],
              'gexc': {'exclude': SUBJECTS[0]},
              'gexc2': {'base': 'gexc', 'exclude': SUBJECTS[-1]}}

    sessions = 'file'


class FileExperimentDefaults(FileExperiment):

    defaults = {'session': 'file',
                'group': 'gsub'}


def test_file_handling():
    "Test MneExperiment with actual files"
    tempdir = TempDir()
    for subject in SUBJECTS:
        sdir = os.path.join(tempdir, 'meg', subject)
        os.makedirs(sdir)

    e = FileExperiment(tempdir)

    eq_(e.get('subject'), SUBJECTS[0])
    eq_([s for s in e.iter(group='all')], SUBJECTS)
    eq_([s for s in e.iter(group='gsub')], SUBJECTS[1:])
    eq_([s for s in e.iter(group='gexc')], SUBJECTS[1:])
    eq_([s for s in e.iter(group='gexc2')], SUBJECTS[1:-1])
    eq_(e.get('subject'), SUBJECTS[1])
    eq_(e.get('subject', group='all'), SUBJECTS[1])
    e.set(SUBJECTS[0])
    eq_(e.get('subject'), SUBJECTS[0])
    eq_(e.get('subject', group='gsub'), SUBJECTS[1])

    e = FileExperimentDefaults(tempdir)
    eq_(e.get('group'), 'gsub')
    eq_(e.get('subject'), SUBJECTS[1])
